#ifndef _MESH_H_
#define _MESH_H_

#define _CRT_SECURE_NO_DEPRECATE //i won't use fopen_s; so, don't warn me

#include <iostream>
#include <vector>
#include <set>
#include <queue>
using namespace std;

#include "Eigen/Dense" //dense matrices
//#include "Eigen/Sparse" //sparse matrices
using namespace Eigen;

#define KDTREE_FOR_CLOSEST_PNTS //use nlogm kd-tree for closest point search instead of the simple nm exhaustive search way b/w 2 pnt sets of sizes n & m
#ifdef KDTREE_FOR_CLOSEST_PNTS
#include "kdtree.h"
#endif

#define EXP							2.718281828f
#define PI							3.141592653f
#define INF							9000000000000000000000000000000.0f //largest float is 3.4x10^38

struct Triangle
{
	//tris[idx] is this tri
	int idx;

	//idx to verts forming this tri
	int v1i, v2i, v3i;
	
	Triangle(int i, int i1, int i2, int i3) : idx(i), v1i(i1), v2i(i2), v3i(i3) {};
};

struct Tetrahedron
{
	//tets[idx] is this tetrahedron
	int idx;

	//idx to tris forming this tetrahedron
	int t1i, t2i, t3i, t4i, //for face-based rendering
		//idx to verts forming this tetrahedron (could've been derived from tris but this is faster)
		v1i, v2i, v3i, v4i; //for vertex-based deformation
	
	//unsigned volume
	float volume; //, surfaceArea, ..

	Tetrahedron(int i, int i1, int i2, int i3, int i4, int j1, int j2, int j3, int j4) : idx(i), t1i(i1), t2i(i2), t3i(i3), t4i(i4), v1i(j1), v2i(j2), v3i(j3), v4i(j4) {};
};

struct Vertex
{
	//verts[idx] is this vert
	int idx,
		closestVertIdx; //closest mesh vertex to this vert during the current ICP iteration

	//current and previous iter's coordinates and deformation direction
	float* coords;

	//idx to vert neighbors of this vert (use set for efficient duplicate entry prevention)
	typedef set< int, std::less< int > > vertNeighborsIS; vertNeighborsIS vertNeighborsSet;
	
	//idx to triangle (tris) and tetrahedron (tets) neighbs
	vector< int > triNeighbors, tetNeighbors;

	//idx to edges incident to this vertex
	vector< int > edgeList, interiorEdgeList; //interior/invisible edges incident to this vert

	Vertex(int i, float* c) : idx(i), coords(c) {};

	bool addVertNeighbor(int vertInd)
	{
		//add vertInd as a vert neighb of this vert if it won't cause a duplication

		pair< vertNeighborsIS::const_iterator, bool > pa;
		//O(log N) search for duplicates by the use of red-black tree (balanced binary search tree)
		pa = vertNeighborsSet.insert(vertInd);
		//pa.second is false if item already exists, and true if insertion succeeds
		return pa.second;
	}
	
	void removeVertNeighbor(int toBeRemoved)
	{
		//size_type erase (const key_type& x): Deletes all the elements matching x. Returns the 
		//number of elements erased. Since a set supports unique keys, erase will always return 1 or 0

		vertNeighborsSet.erase(toBeRemoved);
	}

	void removeTriNeighbor(int t)
	{
//this fnctn could've been used by contract/split/flip as well but i wrote it later; so, only createSphere() uses it

		unsigned int removeIdx;
		for (removeIdx = 0; removeIdx < triNeighbors.size(); removeIdx++)
			if (triNeighbors[removeIdx] == t)
				break;

		if (removeIdx == triNeighbors.size()) { cout << "tri" << t << " not in vert" << idx << ".triNeighbs!\n"; exit(0); }

		triNeighbors.erase(triNeighbors.begin() + removeIdx);
	}
};

struct Edge
{
	//edges[idx] is this edge
	int idx;

	//idx to endpnt verts of this edge
	int v1i, v2i;

	//distance between v1i & v2i
	float length;

	bool interior; //true if this is an interior edge, false if it is on the boundary (i.e. useful for geodesic computations)

	Edge(int i, int i1, int i2, float l, bool in) : idx(i), v1i(i1), v2i(i2), length(l), interior(in) {};
};

class Mesh
{
public:
	//triangular mesh stuff
	vector< Triangle* > tris;
	vector< Vertex* > verts;
	vector< Edge* > edges;
	float minEdgeLen, maxEdgeLen, edgeLenTotal, avgEdgeLen;

	//tetrahedral mesh stuff [there is also a Triangle section in tetmesh files which gives the triangles on the surface, i.e. the visible/outside/boundary ones; i don't load them currently]
	vector< Tetrahedron* > tets;

	Mesh() {};
	
	float ICP(Mesh* dtMesh, bool adaptiveScaling, int nMaxIters);

	bool loadPnt(char* meshFile);

	void resultToFile(char* meshFile);
private:
	int addVertex(float* v);
	int addVertex(float x, float y, float z);
	void addVertexND(float* c);
	int addTriangle(int v1i, int v2i, int v3i);
	int addTriangleTetmesh(int v1i, int v2i, int v3i, bool tetPartInMeshFile);
	void addEdge(int v1i, int v2i, bool interiorEdge = false);
	inline float calcTriArea(int t);
	int* getTrianglesSharedBy(Edge* bigEdge, int v1i = -1, int v2i = -1);
	void addTetrahedron(int v1i, int v2i, int v3i, int v4i);
	int getEdgeSharedBy(int vi, int vj);
};

#endif
